# Copyright 2020 Lawrence Livermore National Security, LLC and other CLIPPy Project Developers.
# See the top-level COPYRIGHT file for details.
#
# SPDX-License-Identifier: MIT

""" This file contains custom Clippy errors. """


class ClippyError(Exception):
    '''
    This is a top-level custom exception for Clippy.
    '''


class ClippyConfigurationError(ClippyError):
    '''
    This error represents a configuration error on user input.
    '''


class ClippyBackendError(ClippyError):
    '''
    This error should be thrown when the backend returns an abend.
    '''


class ClippyValidationError(ClippyError):
    '''
    This error represents a validation error in the inputs to a clippy job.
    '''


class ClippySerializationError(ClippyError):
    '''
    This error should be thrown when clippy object serialization fails
    '''


class ClippyClassInconsistencyError(ClippyError):
    '''
    This error represents a class inconsistency error (name or docstring mismatch).
    '''


class ClippyTypeError(ClippyError):
    '''
    This error represents an error with the type of data being passed to the back end.
    '''


class ClippyInvalidSelectorError(ClippyError):
    '''
    This error represents an error with a selector that is not defined for a given clippy class.
    '''



""" Clippy constants. """

# These should probably not be changed, which is why they're not in config.

# The name of the selector key in the meta.json files.
# these selectors cannot be dropped.
INITIAL_SELECTOR_KEY = 'initial_selectors'

######
# Keys found in output from backend.
######
# The name of the selector key in the json returns.
# If this key exists in the results
# then the selector hierarchy is replaced with its (processed) contents.
SELECTOR_KEY = '_selectors'
# Any internal state kept in the Python frontend.
STATE_KEY = '_state'
# This will reference a string to be printed to stdout by the frontend.
OUTPUT_KEY = 'stdout'
# This key is a boolean - if true, return self (to enable method chaining).
SELF_KEY = 'returns_self'
# key to json entry that holds reference overrides from backend functions.
REFERENCE_KEY = 'references'
# key to json entry that holds return data from backend functions.
RETURN_KEY = 'returns'


# these keys are returned from the _help output when creating methods
# for classes.
METHODNAME_KEY = 'method_name'
DOCSTRING_KEY = 'desc'
ARGS_KEY = 'args'


REAL = 'real'
STRING = 'string'
UINT = 'uint'
INT = 'int'
CLASS_META_FILE = 'meta.json'



# Constants unique to the fs backend.
# the flag to pass for a dry run to make sure syntax is proper

DRY_RUN_FLAG = "--clippy-validate"
# the flag to pass to get detailed help for constructing the class
HELP_FLAG = "--clippy-help"

# Arguments for execution progress
PROGRESS_START_KEY = "progress_start"
PROGRESS_END_KEY = "progress_end"
PROGRESS_INC_KEY = "progress_inc"
PROGRESS_SET_KEY = "progress_set"



"""
    Functions to execute backend programs.
"""

from __future__ import annotations
import json
import logging

import subprocess
from ...clippy_types import AnyDict
from ... import cfg
from ...constants import OUTPUT_KEY
from .constants import (
    DRY_RUN_FLAG,
    HELP_FLAG,
    PROGRESS_INC_KEY,
    PROGRESS_SET_KEY,
    PROGRESS_START_KEY,
    PROGRESS_END_KEY,
)

from ...error import ClippyValidationError, ClippyBackendError
from ..serialization import encode_clippy_json, decode_clippy_json

try:
    from tqdm import tqdm

    _has_tqdm = True
except ImportError:
    _has_tqdm = False


def _stream_exec(
    cmd: list[str],
    submission_dict: AnyDict,
    logger: logging.Logger,
    validate: bool,
) -> tuple[AnyDict | None, str | None]:
    '''
    Internal function.

    Executes the command specified with `execcmd` and
    passes `submission_dict` as JSON via STDIN.

    Logs debug messages with progress.
    Parses the object and returns a dictionary output.
    Returns the process result object.

    This function is used by _run and _validate. All options (pre_cmd and flags) should
    already be set.
    '''

    logger.debug(f'Submission = {submission_dict}')
    # PP support passing objects
    # ~ cmd_stdin = json.dumps(submission_dict)
    cmd_stdin = json.dumps(submission_dict, default=encode_clippy_json)

    logger.debug('Calling %s with input %s', cmd, cmd_stdin)

    d = {}
    stderr = None
    with subprocess.Popen(
        cmd, stdin=subprocess.PIPE, stdout=subprocess.PIPE, stderr=subprocess.PIPE, encoding='utf8'
    ) as proc:
        assert proc.stdin is not None
        assert proc.stdout is not None
        assert proc.stderr is not None

        proc.stdin.write(cmd_stdin + "\n")
        proc.stdin.flush()
        # proc.stdin.close()
        it = iter(proc.stdout.readline, '')
        progress = None
        for line in it:
            d = json.loads(line, object_hook=decode_clippy_json)
            if proc.poll() is not None:  # process terminated; this shouldn't normally happen.
                break
            if _has_tqdm:
                if progress is None:
                    if PROGRESS_START_KEY in d:
                        progress = tqdm() if d[PROGRESS_START_KEY] is None else tqdm(total=d[PROGRESS_START_KEY])
                        # print(f"start, total = {d[PROGRESS_START_KEY]}, {progress.n=}")
                else:
                    if PROGRESS_INC_KEY in d:
                        progress.update(d[PROGRESS_INC_KEY])
                        progress.refresh()
                        # print(f"update {progress.n=}")
                    if PROGRESS_SET_KEY in d:
                        progress.n = d[PROGRESS_SET_KEY]
                        progress.refresh()
                    if PROGRESS_END_KEY in d:
                        progress.close()
                        # print("close")
                        progress = None
            if progress is None:
                if OUTPUT_KEY in d:
                    print(d[OUTPUT_KEY])

        if proc.stderr is not None:
            stderr = "".join(proc.stderr.readlines())
    if progress is not None:
        progress.close()
    if proc.returncode:
        raise (ClippyValidationError(stderr) if validate else ClippyBackendError(stderr))

    if not d:
        return None, stderr
    if stderr:
        logger.debug('Received stderr: %s', stderr)
    logger.debug('run(): final stdout = %s', d)

    return (d, stderr)


def _validate(cmd: str | list[str], dct: AnyDict, logger: logging.Logger) -> tuple[bool, str]:
    '''
    Converts the dictionary dct into a json file and calls executable cmd with the DRY_RUN_FLAG.
    Returns True/False (validation successful) and any stderr.
    '''

    if isinstance(cmd, str):
        cmd = [cmd]

    execcmd = cfg.get('validate_cmd_prefix').split() + cmd + [DRY_RUN_FLAG]
    logger.debug('Validating %s', cmd)

    _, stderr = _stream_exec(execcmd, dct, logger, validate=True)
    return stderr is not None, stderr or ''


def _run(cmd: str | list[str], dct: AnyDict, logger: logging.Logger) -> AnyDict:
    '''
    converts the dictionary dct into a json file and calls executable cmd. Prepends
    cmd_prefix configuration, if any.
    '''

    if isinstance(cmd, str):
        cmd = [cmd]
    execcmd = cfg.get('cmd_prefix').split() + cmd
    logger.debug('Running %s', execcmd)
    # should we do something with stderr?

    output, _ = _stream_exec(execcmd, dct, logger, validate=False)
    return output or {}


def _help(cmd: str | list[str], dct: AnyDict, logger: logging.Logger) -> AnyDict:
    '''
    Retrieves the help output from the clippy command. Prepends validate_cmd_prefix
    if set and appends HELP_FLAG.
    Unlike `_validate()`, does not append DRY_RUN_FLAG, and returns the output.
    '''
    if isinstance(cmd, str):
        cmd = [cmd]
    execcmd = cfg.get('validate_cmd_prefix').split() + cmd + [HELP_FLAG]
    logger.debug('Running %s', execcmd)
    # should we do something with stderr?

    output, _ = _stream_exec(execcmd, dct, logger, validate=True)
    return output or {}



"""
    Utility functions
"""

from .clippy_types import AnyDict
from .error import ClippyInvalidSelectorError
from .constants import SELECTOR_KEY


def flat_dict_to_nested(input_dict: AnyDict) -> AnyDict:
    """input dictionary has dot-delineated keys which are then parsed as subkeys.
    That is: {'a.b.c': 5} becomes {'a': {'_selectors': {'b': {'_selectors': {'c': {'__doc__' = 5}}}}}}
    assuming `SELECTOR_KEY == '_selectors'`
    """

    output_dict: AnyDict = {}
    for k, v in input_dict.items():
        # k is dotted
        if '.' not in k:
            raise ClippyInvalidSelectorError("cannot set top-level selectors")

        *path, last = k.split('.')
        if last.startswith('_'):
            raise ClippyInvalidSelectorError(
                "selectors must not start with an underscore."
            )
        curr_nest = output_dict
        for p in path:
            if p.startswith('_'):
                raise ClippyInvalidSelectorError(
                    "selectors must not start with an underscore."
                )
            curr_nest.setdefault(p, {})
            curr_nest[p].setdefault(SELECTOR_KEY, {})
            curr_nest = curr_nest[p][SELECTOR_KEY]

        curr_nest.setdefault(last, {'__doc__': v})
    return output_dict



# Copyright 2020 Lawrence Livermore National Security, LLC and other CLIPPy Project Developers.
# See the top-level COPYRIGHT file for details.
#
# SPDX-License-Identifier: MIT

""" Holds the expression building code. """

from __future__ import annotations
import json
from typing import Any
from .backends.serialization import ClippySerializable
from . import constants
from .clippy_types import AnyDict


class Expression(ClippySerializable):
    '''A clippy expression is a triple of (operation, operand1, operand2) where
    operation is a string representing an operation (e.g., `"<"` for "less than")
    and operand1 is a Selector object. Operand2 may be a selector object, or any
    constant value. It can also be an expression.
    '''

    # Selectors are currently subclasses of Expression where both op and operand2 are None.
    # This should probably change.

    def __init__(self, op: str | None, o1: Selector, o2: Any):
        super().__init__()
        # if op is None then this expression is just a native Selector held in o1.
        self.op = op
        self.o1 = o1
        self.o2 = o2

    def _express(self, op, o):
        return Expression(op, self, o)

    def __lt__(self, o):
        return self._express("<", o)

    def __le__(self, o):
        return self._express("<=", o)

    def __eq__(self, o):
        return self._express("==", o)

    def __ne__(self, o):
        return self._express("!=", o)

    def __gt__(self, o):
        return self._express(">", o)

    def __ge__(self, o):
        return self._express(">=", o)

    def __add__(self, o):
        return self._express("+", o)

    def __sub__(self, o):
        return self._express("-", o)

    def __mul__(self, o):
        return self._express("*", o)

    def __matmul__(self, o):
        return self._express("@", o)

    def __truediv__(self, o):
        return self._express("/", o)

    def __floordiv__(self, o):
        return self._express("//", o)

    def __mod__(self, o):
        return self._express("%", o)

    def __divmod__(self, o):
        return self._express("divmod", o)

    def __pow__(self, o):
        return self._express("**", o)

    def __lshift__(self, o):
        return self._express("<<", o)

    def __rshift__(self, o):
        return self._express(">>", o)

    def __and__(self, o):
        return self._express("and", o)

    def __xor__(self, o):
        return self._express("^", o)

    def __or__(self, o):
        return self._express("or", o)

    def __contains__(self, o):
        raise NotImplementedError(
            "syntax a in b is not supported. Use b.contains(a) instead."
        )
        # will not work when written as "x in set",
        #   b/c the in-operator always converts the result to bool
        #   https://stackoverflow.com/questions/38542543/functionality-of-python-in-vs-contains
        #   https://bugs.python.org/issue16011
        # return Expression("in", o, self)

    # to be modeled after Pandas' str.contains
    def contains(self, o, regex=False):
        oper = "in" if not regex else "regex"
        return Expression(oper, o, self)

    #        return self._express(oper, o)

    # string and array concatenation
    def cat(self, o):
        return self._express("cat", o)

    def __str__(self):
        return self.to_json()

    def __repr__(self):
        return self.to_json()

    def to_json(self):
        o1 = self.o1
        o2 = self.o2
        return f"{{\"{self.op}\": [{o1},{o2}] }}"

    def to_serial(self):
        o1 = self.o1
        o2 = self.o2
        if hasattr(o1, "to_serial"):
            o1 = o1.to_serial()
        if hasattr(o2, "to_serial"):
            o2 = o2.to_serial()

        return {self.op: [o1, o2]}


class Selector(Expression):  # pylint: disable=abstract-method
    '''A Selector is a subset of `Expression` and represents a single variable.'''

    def __init__(self, parent: Selector | None, name: str, docstr: str):
        super().__init__(
            None, self, None
        )  # op and o2 are None to represent this as a variable.
        self.parent = parent
        self.name = name
        setattr(self, '__doc__', docstr)
        self.fullname: str = (
            self.name if self.parent is None else f"{self.parent.fullname}.{self.name}"
        )
        self.subselectors: set[Selector] = set()

    def __hash__(self):
        return hash(self.fullname)

    def to_dict(self):
        return {'var': self.fullname}

    def hierarchy(self, acc: list[tuple[str, str]] | None = None):
        if acc is None:
            acc = []
        acc.append((self.fullname, self.__doc__ or ''))
        for subsel in self.subselectors:
            subsel.hierarchy(acc)
        return acc

    def describe(self):
        hier = self.hierarchy()
        maxlen = max((len(sub_desc[0]) for sub_desc in hier))
        return '\n'.join(
            f'{sub_desc[0]:<{maxlen+2}} {sub_desc[1]}' for sub_desc in hier
        )

    def __str__(self):
        return repr(self.to_dict())

    def __repr__(self):
        return repr(self.to_dict())

    def to_json(self):
        return json.dumps(self.to_dict())

    def to_serial(self):
        return {"var": self.fullname}

    def _express(self, op, o):
        return Expression(op, self, o)

    def _add_subselector(self, name: str, docstr: str):
        '''add a subselector to this selector'''
        subsel = Selector(self, name, docstr)
        setattr(self, name, subsel)
        self.subselectors.add(subsel)

    def _del_subselector(self, name: str):
        delattr(self, name)
        self.subselectors.remove(getattr(self, name))

    def _clear_subselectors(self):
        '''removes all subselectors'''
        for subsel in self.subselectors:
            delattr(self, subsel.name)
        self.subselectors = set()

    def _import_from_dict(self, d: AnyDict, merge: bool = False):
        '''Imports subselectors from a dictionary.
        If `merge = True`, do not clear subselectors first.
        '''

        # clear all children
        if not merge:
            self._clear_subselectors()

        for name, subdict in d.get(constants.SELECTOR_KEY, {}).items():
            docstr = subdict.get('__doc__', '')
            self._add_subselector(name, docstr)
            getattr(self, name)._import_from_dict(subdict)



import pytest
import sys

sys.path.append("src")

import jsonlogic as jl
import clippy
from clippy.error import ClippyValidationError, ClippyInvalidSelectorError

import logging

clippy.logger.setLevel(logging.WARN)
logging.getLogger().setLevel(logging.WARN)


@pytest.fixture()
def testbag():
    return clippy.TestBag()


@pytest.fixture()
def testset():
    return clippy.TestSet()


@pytest.fixture()
def testfun():
    return clippy.TestFunctions()


@pytest.fixture()
def testsel():
    return clippy.TestSelector()


def test_imports():
    assert "TestBag" in clippy.__dict__


def test_bag(testbag):

    testbag.insert(41)
    assert testbag.size() == 1
    testbag.insert(42)
    assert testbag.size() == 2
    testbag.insert(41)
    assert testbag.size() == 3
    testbag.remove(41)
    assert testbag.size() == 2
    testbag.remove(99)
    assert testbag.size() == 2



def test_clippy_call_with_string(testfun):
    assert testfun.call_with_string("Seth") == "Howdy, Seth"
    with pytest.raises(ClippyValidationError):
        testfun.call_with_string()


def test_expression_gt_gte(testbag):
    testbag.insert(10).insert(41).insert(42).insert(50).insert(51).insert(52)
    assert testbag.size() == 6
    testbag.remove_if(testbag.value > 51)
    assert testbag.size() == 5
    testbag.remove_if(testbag.value >= 50)
    assert testbag.size() == 3
    testbag.remove_if(testbag.value >= 99)
    assert testbag.size() == 3


def test_expression_lt_lte(testbag):
    testbag.insert(10).insert(41).insert(42).insert(50).insert(51).insert(52)
    testbag.remove_if(testbag.value < 42)
    assert testbag.size() == 4
    testbag.remove_if(testbag.value <= 51)
    assert testbag.size() == 1


def test_expression_eq_neq(testbag):
    testbag.insert(10).insert(11).insert(12)
    assert testbag.size() == 3
    testbag.remove_if(testbag.value != 11)
    assert testbag.size() == 1
    testbag.remove_if(testbag.value == 11)
    assert testbag.size() == 0


def test_expresssion_add(testbag):
    testbag.insert(10).insert(41).insert(42).insert(50).insert(51).insert(52)
    testbag.remove_if(testbag.value + 30 > 70)
    assert testbag.size() == 1


def test_expression_sub(testbag):
    testbag.insert(10).insert(41).insert(42).insert(50).insert(51).insert(52)
    testbag.remove_if(testbag.value - 30 > 0)
    assert testbag.size() == 1


def test_expression_mul_div(testbag):
    testbag.insert(10).insert(41).insert(42).insert(50).insert(51).insert(52)
    testbag.remove_if(testbag.value * 2 / 4 > 10)
    assert testbag.size() == 1


def test_expression_or(testbag):
    testbag.insert(10).insert(41).insert(42).insert(50).insert(51).insert(52)
    testbag.remove_if((testbag.value < 41) | (testbag.value > 49))
    assert testbag.size() == 2  # 41, 42


def test_expression_and(testbag):
    testbag.insert(10).insert(41).insert(42).insert(50).insert(51).insert(52)
    testbag.remove_if((testbag.value > 40) & (testbag.value < 50))
    assert testbag.size() == 4  # 10, 50, 51, 52


# TODO: not yet implemented
# def test_expression_floordiv(testbag):
#     testbag.insert(10).insert(41).insert(42).insert(50).insert(51).insert(52)
#     testbag.remove_if(testbag.value * 2 // 4.2 > 10)
#     assert testbag.size() == 1


def test_expression_mod(testbag):
    testbag.insert(10).insert(41).insert(42).insert(50).insert(51).insert(52)
    testbag.remove_if(testbag.value % 2 == 0)
    assert testbag.size() == 2


# TODO: not yet implemented
# def test_expression_pow(testbag):
#     testbag.insert(10).insert(41).insert(42).insert(50).insert(51).insert(52)
#     testbag.remove_if(testbag.value**2 > 1000)
#     assert testbag.size() == 2



def test_clippy_returns_int(testfun):
    assert testfun.returns_int() == 42


def test_clippy_returns_string(testfun):
    assert testfun.returns_string() == "asdf"


def test_clippy_returns_bool(testfun):
    assert testfun.returns_bool()


def test_clippy_returns_dict(testfun):
    d = testfun.returns_dict()
    assert len(d) == 3
    assert d.get("a") == 1
    assert d.get("b") == 2
    assert d.get("c") == 3


def test_clippy_returns_vec_int(testfun):
    assert testfun.returns_vec_int() == [0, 1, 2, 3, 4, 5]


def test_clippy_returns_optional_string(testfun):
    assert testfun.call_with_optional_string() == "Howdy, World"
    assert testfun.call_with_optional_string(name="Seth") == "Howdy, Seth"


def test_selectors(testsel):
    assert hasattr(testsel, "nodes")

    testsel.add(testsel.nodes, "b", desc="docstring for nodes.b").add(
        testsel.nodes.b, "c", desc="docstring for nodes.b.c"
    )
    assert hasattr(testsel.nodes, "b")
    assert hasattr(testsel.nodes.b, "c")
    assert testsel.nodes.b.__doc__ == "docstring for nodes.b"
    assert testsel.nodes.b.c.__doc__ == "docstring for nodes.b.c"

    assert isinstance(testsel.nodes.b, jl.Variable)
    assert isinstance(testsel.nodes.b.c, jl.Variable)

    with pytest.raises(ClippyInvalidSelectorError):
        testsel.add(testsel.nodes, "_bad", desc="this is a bad selector name")

    # with pytest.raises(ClippyInvalidSelectorError):
    #     testsel.add(testsel, 'bad', desc="this is a top-level selector")
